#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <hpic.h>

int main(int argc, char *argv[])
{
  char ch = '-';
  size_t i, j, k;

  char first[HPIC_STRNL];
  char second[HPIC_STRNL];
  int intersect = 1;

  size_t fnside;
  size_t snside;
  int forder;
  int sorder;
  int fcoord;
  int scoord;
  int ftype;
  int stype;

  char command[HPIC_STRNL];
  char comment[HPIC_STRNL];
  char screator[HPIC_STRNL];
  char sextname[HPIC_STRNL];
  char fcreator[HPIC_STRNL];
  char fextname[HPIC_STRNL];

  size_t snmaps;
  size_t fnmaps;
  size_t fmap = 0;
  size_t smap = 0;
  hpic_int *fpix = NULL;
  hpic_fltarr *fdata = NULL;
  hpic_int *fhits = NULL;
  hpic_float *ferrs = NULL;
  hpic_keys *fkeys = NULL;
  hpic_float *ftemp = NULL;

  hpic_int *spix = NULL;
  hpic_fltarr *sdata = NULL;
  hpic_int *shits = NULL;
  hpic_float *serrs = NULL;
  hpic_keys *skeys = NULL;
  hpic_float *stemp = NULL;

  int temp;

  if (argc == 1) {
    printf("\n");
    printf("<******************************************************>\n");
    printf("Usage:  hpic_sum\n");
    printf("        -first <first FITS file>\n");
    printf("        -fmap <map number to use in first file>\n");
    printf("        -second <second FITS file>\n");
    printf("        -smap <map number to use in second file>\n");
    printf("        -union (keep union of maps)\n");
    printf("<******************************************************>\n\n");
    exit(1);
  }

  /* parse command line */

  for (i = 0; i < argc; i++) {
    if (argv[i][0] == ch) {
      if (!strcmp(argv[i], "-first")) {
        strncpy(first, argv[i + 1], HPIC_STRNL);
        i++;
      }
      if (!strcmp(argv[i], "-fmap")) {
        fmap = (size_t) atoi(argv[i + 1]);
        i++;
      }
      if (!strcmp(argv[i], "-smap")) {
        smap = (size_t) atoi(argv[i + 1]);
        i++;
      }
      if (!strcmp(argv[i], "-second")) {
        strncpy(second, argv[i + 1], HPIC_STRNL);
        i++;
      }
      if (!strcmp(argv[i], "-union")) {
        intersect = 0;
      }
    }
  }

  /* read second FITS file */

  if (hpic_fits_map_test(second, &snside, &sorder, &scoord, &stype, &snmaps)) {
    skeys = hpic_keys_alloc();
    sdata = hpic_fltarr_alloc(snmaps);
    for (i = 0; i < snmaps; i++) {
      stemp = hpic_float_alloc(snside, sorder, scoord, HPIC_AUTO);
      hpic_fltarr_set(sdata, i, stemp);
    }
    if (stype == HPIC_FITS_FULL) {
      hpic_fits_full_read(second, screator, sextname, sdata, skeys);
    } else {
      serrs = hpic_float_alloc(snside, sorder, scoord, HPIC_AUTO);
      shits = hpic_int_alloc(snside, sorder, scoord, HPIC_AUTO);
      spix = hpic_int_alloc(snside, sorder, scoord, HPIC_AUTO);
      hpic_fits_cut_read(second, screator, sextname, spix, shits, serrs,
                         sdata, skeys);
      hpic_int_free(spix);
      hpic_int_free(shits);
      hpic_float_free(serrs);
    }
    hpic_keys_free(skeys);
    stemp = hpic_fltarr_get(sdata, smap);
  } else {
    fprintf(stderr,
            "hpic_sum:  second FITS file is not a recognized map type\n");
    exit(1);
  }

  /* read first file, subtract second, and write to original file */

  if (hpic_fits_map_test(first, &fnside, &forder, &fcoord, &ftype, &fnmaps)) {
    fkeys = hpic_keys_alloc();
    fdata = hpic_fltarr_alloc(fnmaps);
    for (i = 0; i < fnmaps; i++) {
      ftemp = hpic_float_alloc(fnside, forder, fcoord, HPIC_AUTO);
      hpic_fltarr_set(fdata, i, ftemp);
    }
    if (ftype == HPIC_FITS_FULL) {
      hpic_fits_full_read(first, fcreator, fextname, fdata, fkeys);
      ftemp = hpic_fltarr_get(fdata, fmap);
      if (intersect) {
        hpic_float_add(ftemp, stemp, HPIC_INTERSECT);
      } else {
        hpic_float_add(ftemp, stemp, HPIC_UNION);
      }
      snprintf(command, HPIC_STRNL, "mv %s %s.bak", first, first);
      system(command);
      snprintf(comment, HPIC_STRNL, "added map %lu of file %s to map %lu",
               smap, second, fmap);
      strcpy(fcreator, "hpic_sum");
      hpic_fits_full_write(first, fcreator, fextname, comment, fdata, fkeys);
    } else {
      ferrs = hpic_float_alloc(fnside, forder, fcoord, HPIC_AUTO);
      fhits = hpic_int_alloc(fnside, forder, fcoord, HPIC_AUTO);
      fpix = hpic_int_alloc(fnside, forder, fcoord, HPIC_AUTO);
      hpic_fits_cut_read(first, fcreator, fextname, fpix, fhits, ferrs, fdata,
                         fkeys);
      ftemp = hpic_fltarr_get(fdata, fmap);
      if (intersect) {
        hpic_float_add(ftemp, stemp, HPIC_INTERSECT);
      } else {
        hpic_float_add(ftemp, stemp, HPIC_UNION);
      }
      snprintf(command, HPIC_STRNL, "mv %s %s.bak", first, first);
      system(command);
      snprintf(comment, HPIC_STRNL, "added map %lu of file %s to map %lu",
               smap, second, fmap);
      strcpy(fcreator, "hpic_sum");
      hpic_int_setall(fpix, HPIC_INT_NULL);
      for (i = 0; i < hpic_int_npix_get(fpix); i++) {
        for (j = 0; j < fnmaps; j++) {
          ftemp = hpic_fltarr_get(fdata, j);
          if (!hpic_is_fnull(hpic_float_get(ftemp, i))) {
            hpic_int_set(fpix, i, (int)i);
          }
        }
      }
      hpic_fits_cut_write(first, fcreator, fextname, comment, fpix, fhits,
                          ferrs, fdata, fkeys);
      hpic_int_free(fpix);
      hpic_int_free(fhits);
      hpic_float_free(ferrs);
    }
    hpic_keys_free(fkeys);
    for (i = 0; i < fnmaps; i++) {
      ftemp = hpic_fltarr_get(fdata, i);
      hpic_float_free(ftemp);
    }
    hpic_fltarr_free(fdata);
  } else {
    fprintf(stderr,
            "hpic_sum:  first FITS file is not a recognized map type\n");
    exit(1);
  }

  for (i = 0; i < snmaps; i++) {
    stemp = hpic_fltarr_get(sdata, i);
    hpic_float_free(stemp);
  }
  hpic_fltarr_free(sdata);

  return 0;
}
